package ee.ut.cs.gitlab.maikov.matrix;

import java.util.Arrays;

public class Matrix {

    protected double[] elements;
    protected int rows;
    protected int cols;

    public Matrix(int rows, int columns) {
        if (rows <= 0 || columns <= 0) {
            throw new IllegalArgumentException("Rows or columns must be greater than zero!");
        }

        this.rows = rows;
        this.cols = columns;
        this.elements = new double[rows*columns];
        Arrays.fill(this.elements, 0.0);
    }

    public Matrix(int rows, int columns, double[] elements) {
        if (rows <= 0 || columns <= 0) {
            throw new IllegalArgumentException("Rows or columns must be greater than zero!");
        }

        if (elements.length != rows * columns) {
            throw new IllegalArgumentException("Elements must be the same length as rows and columns!");
        }

        this.rows = rows;
        this.cols = columns;
        this.elements = Arrays.copyOf(elements, elements.length);
    }

    protected int getIndex(int row, int col) {
        return row*cols + col;
    }

    public void set(int row, int col, double value) {
        this.elements[this.getIndex(row, col)] = value;
    }

    public double get(int row, int col) {
        return this.elements[this.getIndex(row, col)];
    }

    public void add(Matrix other) {
        if (rows != other.rows || cols != other.cols) {
            throw new IllegalArgumentException("Rows and columns must be the same " +
                    "between matrices for addition!");
        }

        for (int i = 0; i < elements.length; i++) {
            this.elements[i] += other.elements[i];
        }
    }

    public void multiply(double factor) {
        for (int i = 0; i < elements.length; i++) {
            elements[i] *= factor;
        }
    }

    public void multiply(Matrix other) {
        if (this.cols != other.rows) {
            throw new IllegalArgumentException("Matrix multiplication requires mxn and nxp matrices.");
        }

        Matrix result = new Matrix(this.rows, other.cols);

        for (int i = 0; i < this.rows; i++) {
            for (int j = 0; j < other.cols; j++) {
                double sum = 0.0;
                for (int k = 0; k < this.cols; k++) {
                    sum += this.elements[this.getIndex(i, k)] * other.elements[other.getIndex(k, j)];
                }
                result.elements[result.getIndex(i, j)] = sum;
            }
        }

        this.rows = result.rows;
        this.cols = result.cols;
        this.elements = result.elements;
    }

    public Matrix transpose() {
        Matrix result = new Matrix(this.cols, this.rows);

        for (int i = 0; i < this.rows; i++) {
            for (int j = 0; j < this.cols; j++) {
                result.set(j, i, this.get(i, j));
            }
        }

        return result;
    }

    public int getRows() {
        return rows;
    }

    public int getCols() {
        return cols;
    }

    public double[] getElements() {
        return elements;
    }

    @Override
    public String toString() {
        StringBuilder stringBuilder = new StringBuilder();
        for (int row = 0; row < rows; row++) {
            stringBuilder.append('[');
            for (int col = 0; col < cols; col++) {
                stringBuilder.append(this.get(row, col));
                if (col != cols-1) {
                    stringBuilder.append(',').append(' ');
                }
            }
            stringBuilder.append(']').append('\n');
        }
        return stringBuilder.toString();
    }

    public static Matrix clone(Matrix other) {
        return new Matrix(other.getRows(), other.getCols(), other.getElements());
    }
}
